/*
 * The contents of this file are subject to the Open Software License
 * Version 3.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/osl-3.0.txt
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 */
package org.mulgara.connection;

import java.net.URI;

import org.mulgara.query.Answer;
import org.mulgara.query.Query;
import org.mulgara.query.QueryException;
import org.mulgara.query.TuplesException;
import org.mulgara.server.Session;
import org.mulgara.query.operation.*;

/**
 * This class abstracts connections to a server, holding any information relevant to that
 * connection.  For those operations that are to be performed on a server, this interface
 * is used to send the operations.  Other operations can be kept local, but the mechanism
 * appears the same to the user, thereby abstracting away the interaction that each command
 * has with servers.
 *
 * While Connections can be created with the normal constructor, it may be preferable to use
 * an instance of {@link org.mulgara.connection.ConnectionFactory} in order to cache connections
 * based on the server URI.
 *
 * {@link org.mulgara.query.operation.Command}s to be issued may be executed with a Connection
 * as a parameter, or can be passed to a Connection.  The appropriate use depends on the usage.
 * When creating commands in code, the preferred idiom is to pass to a Connection.  For instance:
 * <pre><code>
 *  ConnectionFactory factory = new ConnectionFactory();
 *  Connection conn = factory.newConnection(URI.create("rmi://localhost/server1"));
 *
 *  URI graph = URI.create("rmi://localhost/server1#graph");
 *  Command create = new CreateGraph(graph);
 *  Command load = new Load(new File("data.rdf").toURI(), graph, false);
 *  Command query = new TqlInterpreter().parseCommand(queryString);
 *
 *  conn.execute(create);
 *  conn.execute(load);
 *  Answer answer = conn.execute(query);
 * </code></pre>
 * 
 * This mechanism has the advantage of returning the appropriate type for each type of command
 * For instance, {@link Query} commands return an {@link Answer}, while most other commands
 * return a status string.
 *
 * Alternatively, if the commands are being generated by a query language parser, then the command
 * operation should be picked up via polymorphism by calling execute on the command instead.
 * <pre><code>
 *  ConnectionFactory factory = new ConnectionFactory();
 *  Connection conn = factory.newConnection(URI.create("rmi://localhost/server1"));
 *  
 *  Interpreter interpreter = new TqlInterpreter();
 *  Command cmd = interpreter.parseCommand(commandString);
 *  
 *  Object result = cmd.execute(conn);
 * </code></pre>
 * 
 * @created 2007-09-25
 * @author Paul Gearon
 * @copyright &copy; 2007 <a href="mailto:pgearon@users.sourceforge.net">Paul Gearon</a>
 * @licence <a href="{@docRoot}/../../LICENCE.txt">Open Software License v3.0</a>
 */
public interface Connection {

  /**
   * Give login credentials and security domain to the current session.  This should only be needed
   * once since the session does not change.
   * @param securityDomainUri The security domain for the login.
   * @param user The username.
   * @param password The password for the given username.
   */
  public void setCredentials(URI securityDomainUri, String user, char[] password);
  

  /**
   * Give login credentials for the current security domain to the current session.
   * This should only be needed
   * once since the session does not change.
   * @param user The username.
   * @param password The password for the given username.
   */
  public void setCredentials(String user, char[] password);


  /**
   * @return the session
   */
  public Session getSession();


  /**
   * Starts and commits transactions on this connection, by turning the autocommit
   * flag on and off. 
   * @param autocommit <code>true</code> if the flag is to be on.
   * @throws QueryException The session could not change state.
   */
  public void setAutoCommit(boolean autoCommit) throws QueryException;


  /**
   * @return the autoCommit value
   */
  public boolean getAutoCommit();


  /**
   * Closes the current connection.
   */
  public void close() throws QueryException;

  // Central execution of Command operations

  /**
   * Generic command execution method.
   * @param The command to execute.
   * @return A status message
   * @throws Exception A general exception catchall
   */
  public String execute(Command cmd) throws Exception;

  /**
   * Loads data from a file or URL
   * @param cmd The command to load the data
   * @return The number of loaded statements
   */
  public Long execute(Load cmd) throws QueryException;

  /**
   * Issues a query on the connection.
   * @param cmd The command to issue the query.
   * @return An Answer with the query results.
   */
  public Answer execute(Query cmd) throws QueryException, TuplesException;

}